# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.modules
"""
Askey-Wilson Algebras

AUTHORS:

- Travis Scrimshaw (2018-08): initial version
"""

# ****************************************************************************
#  Copyright (C) 2018 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.categories.algebras import Algebras
from sage.categories.cartesian_product import cartesian_product
from sage.categories.rings import Rings
from sage.combinat.free_module import CombinatorialFreeModule
from sage.modules.with_basis.morphism import ModuleMorphismByLinearity
from sage.sets.family import Family
from sage.rings.polynomial.laurent_polynomial_ring import LaurentPolynomialRing
from sage.sets.non_negative_integers import NonNegativeIntegers


class AskeyWilsonAlgebra(CombinatorialFreeModule):
    r"""
    The (universal) Askey-Wilson algebra.

    Let `R` be a commutative ring. The *universal Askey-Wilson* algebra
    is an associative unital algebra `\Delta_q` over `R[q,q^-1]` given
    by the generators `A, B, C, \alpha, \beta, \gamma` that satisfy the
    following relations:

    .. MATH::

        \begin{aligned}
        (q-q^{-1}) \alpha &= (q^2-q^{-2}) A + qBC - q^{-1}CB, \\
        (q-q^{-1}) \beta  &= (q^2-q^{-2}) B + qCA - q^{-1}AC, \\
        (q-q^{-1}) \gamma &= (q^2-q^{-2}) C + qAB - q^{-1}BA.
        \end{aligned}

    The universal Askey-Wilson contains a
    :meth:`Casimir element <casimir_element>` `\Omega`, and the elements
    `\alpha`, `\beta`, `\gamma`, `\Omega` generate the center of `\Delta_q`,
    which is isomorphic to the polynomial ring
    `(R[q,q^-1])[\alpha,\beta,\gamma,\Omega]` (assuming `q` is not a root
    of unity). Furthermore, the relations imply that `\Delta_q` has a basis
    given by monomials `A^i B^j C^k \alpha^r \beta^s \gamma^t`, where
    `i, j, k, r, s, t \in \ZZ_{\geq 0}`.

    The universal Askey-Wilson algebra also admits a faithful action
    of `PSL_2(\ZZ)` given by the automorphisms `\rho`
    (:meth:`permutation_automorphism`):

    .. MATH::

        A \mapsto B \mapsto C \mapsto A,
        \qquad
        \alpha \mapsto \beta \mapsto \gamma \mapsto \alpha.

    and `\sigma` (:meth:`reflection_automorphism`):

    .. MATH::

        A \mapsto B \mapsto A,
        C \mapsto C + \frac{AB - BA}{q-q^{-1}},
        \qquad
        \alpha \mapsto \beta \mapsto \alpha,
        \gamma \mapsto \gamma.

    Note that `\rho^3 = \sigma^2 = 1` and

    .. MATH::

        \sigma(C) = C - q AB - (1+q^2) C + q \gamma
                  = C - q AB - q^2 C + q \gamma.

    The Askey-Wilson `AW_q(a,b,c)` algebra is a specialization of the
    universal Askey-Wilson algebra by `\alpha = a`, \beta = b`,
    `\gamma = c`, where `a,b,c \in R`. `AW_q(a,b,c)` was first introduced
    by [Zhedanov1991]_ to describe the Askey-Wilson polynomials. The
    Askey-Wilson algebra has a central extension of `\Delta_q`.

    INPUT:

    - ``R`` -- a commutative ring
    - ``q`` -- (optional) the parameter `q`; must be invertible in ``R``

    If ``q`` is not specified, then ``R`` is taken to be the base
    ring of a Laurent polynomial ring with variable `q`. Otherwise
    the element ``q`` must be an element of ``R``.

    .. NOTE::

        No check is performed to ensure ``q`` is not a root of unity,
        which may lead to violations of the results in [Terwilliger2011]_.

    EXAMPLES:

    We create the universal Askey-Wilson algebra and check
    the defining relations::

        sage: AW = algebras.AskeyWilson(QQ)
        sage: AW.inject_variables()
        Defining A, B, C, a, b, g
        sage: q = AW.q()
        sage: (q^2-q^-2)*A + q*B*C - q^-1*C*B == (q-q^-1)*a
        True
        sage: (q^2-q^-2)*B + q*C*A - q^-1*A*C == (q-q^-1)*b
        True
        sage: (q^2-q^-2)*C + q*A*B - q^-1*B*A == (q-q^-1)*g
        True

    Next, we perform some computations::

        sage: C * A
        (q^-2)*A*C + (q^-3-q)*B - (q^-2-1)*b
        sage: B^2 * g^2 * A
        q^4*A*B^2*g^2 - (q^-1-q^7)*B*C*g^2 + (1-q^4)*B*g^3
         + (1-2*q^4+q^8)*A*g^2 - (q-q^3-q^5+q^7)*a*g^2
        sage: (B^3 - A) * (C^2 + q*A*B)
        q^7*A*B^4 + B^3*C^2 - (q^2-q^14)*B^3*C + (q-q^7)*B^3*g - q*A^2*B
         + (3*q^3-4*q^7+q^19)*A*B^2 - A*C^2 - (1-q^6-q^8+q^14)*B^2*a
         - (q^-2-3*q^6+3*q^14-q^22)*B*C
         + (q^-1+q-3*q^3-q^5+2*q^7-q^9+q^13+q^15-q^19)*B*g
         + (2*q^-1-6*q^3+5*q^7-2*q^19+q^23)*A
         - (2-2*q^2-4*q^4+4*q^6+q^8-q^10+q^12-q^14+q^16-q^18-q^20+q^22)*a

    We check the elements `\alpha`, `\beta`, and `\gamma`
    are in the center::

        sage: all(x * gen == gen * x for gen in AW.algebra_generators() for x in [a,b,g])
        True

    We verify that the :meth:`Casimir element <casimir_element>`
    is in the center::

        sage: Omega = AW.casimir_element()
        sage: all(x * Omega == Omega * x for x in [A,B,C])
        True

        sage: x = AW.an_element()
        sage: O2 = Omega^2
        sage: x * O2 == O2 * x
        True

    We prove Lemma 2.1 in [Terwilliger2011]_::

        sage: (q^2-q^-2) * C == (q-q^-1) * g - (q*A*B - q^-1*B*A)
        True
        sage: (q-q^-1) * (q^2-q^-2) * a == (B^2*A - (q^2+q^-2)*B*A*B + A*B^2
        ....:                               + (q^2-q^-2)^2*A + (q-q^-1)^2*B*g)
        True
        sage: (q-q^-1) * (q^2-q^-2) * b == (A^2*B - (q^2+q^-2)*A*B*A + B*A^2
        ....:                               + (q^2-q^-2)^2*B + (q-q^-1)^2*A*g)
        True

    We prove Theorem 2.2 in [Terwilliger2011]_::

        sage: q3 = q^-2 + 1 + q^2
        sage: A^3*B - q3*A^2*B*A + q3*A*B*A^2 - B*A^3 == -(q^2-q^-2)^2 * (A*B - B*A)
        True
        sage: B^3*A - q3*B^2*A*B + q3*B*A*B^2 - A*B^3 == -(q^2-q^-2)^2 * (B*A - A*B)
        True
        sage: (A^2*B^2 - B^2*A^2 + (q^2+q^-2)*(B*A*B*A-A*B*A*B)
        ....:  == -(q^1-q^-1)^2 * (A*B - B*A) * g)
        True

    We construct an Askey-Wilson algebra over `\GF{5}` at `q=2`::

        sage: AW = algebras.AskeyWilson(GF(5), q=2)
        sage: A,B,C,a,b,g = AW.algebra_generators()
        sage: q = AW.q()
        sage: Omega = AW.casimir_element()

        sage: B * A
        4*A*B + 2*g
        sage: C * A
        4*A*C + 2*b
        sage: C * B
        4*B*C + 2*a
        sage: Omega^2
        A^2*B^2*C^2 + A^3*B*C + A*B^3*C + A*B*C^3 + A^4 + 4*A^3*a
         + 2*A^2*B^2 + A^2*B*b + 2*A^2*C^2 + 4*A^2*C*g + 4*A^2*a^2
         + 4*A*B^2*a + 4*A*C^2*a + B^4 + B^3*b + 2*B^2*C^2 + 4*B^2*C*g
         + 4*B^2*b^2 + B*C^2*b + C^4 + 4*C^3*g + 4*C^2*g^2 + 2*a*b*g

        sage: (q^2-q^-2)*A + q*B*C - q^-1*C*B == (q-q^-1)*a
        True
        sage: (q^2-q^-2)*B + q*C*A - q^-1*A*C == (q-q^-1)*b
        True
        sage: (q^2-q^-2)*C + q*A*B - q^-1*B*A == (q-q^-1)*g
        True
        sage: all(x * Omega == Omega * x for x in [A,B,C])
        True

    REFERENCES:

    - [Terwilliger2011]_
    """
    @staticmethod
    def __classcall_private__(cls, R, q=None):
        r"""
        Normalize input to ensure a unique representation.

        TESTS::

            sage: R.<q> = LaurentPolynomialRing(QQ)
            sage: AW1 = algebras.AskeyWilson(QQ)
            sage: AW2 = algebras.AskeyWilson(R, q)
            sage: AW1 is AW2
            True

            sage: AW = algebras.AskeyWilson(ZZ, 0)
            Traceback (most recent call last):
            ...
            ValueError: q cannot be 0

            sage: AW = algebras.AskeyWilson(ZZ, 3)
            Traceback (most recent call last):
            ...
            ValueError: q=3 is not invertible in Integer Ring
        """
        if q is None:
            R = LaurentPolynomialRing(R, 'q')
            q = R.gen()
        else:
            q = R(q)
        if q == 0:
            raise ValueError("q cannot be 0")
        if 1 / q not in R:
            raise ValueError("q={} is not invertible in {}".format(q, R))
        if R not in Rings().Commutative():
            raise ValueError("{} is not a commutative ring".format(R))
        return super().__classcall__(cls, R, q)

    def __init__(self, R, q) -> None:
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: TestSuite(AW).run()  # long time
        """
        self._q = q
        cat = Algebras(Rings().Commutative()).WithBasis()
        indices = cartesian_product([NonNegativeIntegers()]*6)
        CombinatorialFreeModule.__init__(self, R, indices, prefix='AW',
                                         sorting_key=_basis_key,
                                         sorting_reverse=True,
                                         category=cat)
        self._assign_names('A,B,C,a,b,g')

    def _repr_term(self, t) -> str:
        r"""
        Return a string representation of the basis element indexed by ``t``.

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: AW._repr_term((0,0,0,0,0,0))
            '1'
            sage: AW._repr_term((5,1,2,3,7,2))
            'A^5*B*C^2*a^3*b^7*g^2'
            sage: AW._repr_term((0,1,0,3,7,2))
            'B*a^3*b^7*g^2'
        """
        def exp(l, e):
            if e == 0:
                return ''
            if e == 1:
                return '*' + l
            return '*' + l + '^{}'.format(e)
        ret = ''.join(exp(l, e) for l, e in zip('ABCabg', t))
        if not ret:
            return '1'
        if ret[0] == '*':
            ret = ret[1:]
        return ret

    def _latex_term(self, t) -> str:
        r"""
        Return a latex representation of the basis element indexed by ``t``.

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: AW._latex_term((0,0,0,0,0,0))
            '1'
            sage: AW._latex_term((5,1,2,3,7,2))
            'A^{5}BC^{2}\\alpha^{3}\\beta^{7}\\gamma^{2}'
            sage: AW._latex_term((0,1,0,3,7,2))
            'B\\alpha^{3}\\beta^{7}\\gamma^{2}'
        """
        if sum(t) == 0:
            return '1'

        def exp(l, e):
            if e == 0:
                return ''
            if e == 1:
                return l
            return l + '^{{{}}}'.format(e)
        var_names = ['A', 'B', 'C', '\\alpha', '\\beta', '\\gamma']
        return ''.join(exp(l, e) for l, e in zip(var_names, t))

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: algebras.AskeyWilson(QQ)
            Askey-Wilon algebra with q=q over
             Univariate Laurent Polynomial Ring in q over Rational Field
        """
        return "Askey-Wilon algebra with q={} over {}".format(self._q, self.base_ring())

    @cached_method
    def algebra_generators(self):
        r"""
        Return the algebra generators of ``self``.

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: G = AW.algebra_generators()
            sage: G['A']
            A
            sage: G['a']
            a
            sage: list(G)
            [A, B, C, a, b, g]
        """
        A = self.variable_names()

        def build_monomial(g):
            exp = [0] * 6
            exp[A.index(g)] = 1
            return self.monomial(self._indices(exp))
        return Family(A, build_monomial)

    @cached_method
    def gens(self) -> tuple:
        r"""
        Return the generators of ``self``.

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: AW.gens()
            (A, B, C, a, b, g)
        """
        return tuple(self.algebra_generators())

    @cached_method
    def one_basis(self):
        r"""
        Return the index of the basis element `1` of ``self``.

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: AW.one_basis()
            (0, 0, 0, 0, 0, 0)
        """
        return self._indices([0]*6)

    def q(self):
        r"""
        Return the parameter `q` of ``self``.

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: q = AW.q()
            sage: q
            q
            sage: q.parent()
            Univariate Laurent Polynomial Ring in q over Rational Field
        """
        return self._q

    @cached_method
    def _an_element_(self):
        r"""
        Return an element of ``self``.

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: AW.an_element()
            (q^-3+3+2*q+q^2)*a*b*g^3 + q*A*C^2*b + 3*q^2*B*a^2*g + A
        """
        q = self._q
        I = self._indices
        R = self.base_ring()
        elt = {I((1,0,0,0,0,0)): R(1),
               I((1,0,2,0,1,0)): R.an_element(),
               I((0,1,0,2,0,1)): q**2 * R(3),
               I((0,0,0,1,1,3)): q**-3 + R(3) + R(2)*q + q**2}
        return self.element_class(self, elt)

    def some_elements(self):
        r"""
        Return some elements of ``self``.

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: AW.some_elements()
            (A, B, C, a, b, g, 1,
             (q^-3+3+2*q+q^2)*a*b*g^3 + q*A*C^2*b + 3*q^2*B*a^2*g + A,
             q*A*B*C + q^2*A^2 - q*A*a + (q^-2)*B^2 - (q^-1)*B*b + q^2*C^2 - q*C*g)
        """
        return self.gens() + (self.one(), self.an_element(), self.casimir_element())

    @cached_method
    def casimir_element(self):
        r"""
        Return the Casimir element of ``self``.

        The Casimir element of the Askey-Wilson algebra `\Delta_q` is

        .. MATH::

            \Omega = q ABC + q^2 A^2 + q^{-2} B^2 + q^2 C^2
                     - q A\alpha - q^{-1} B\beta - q C\gamma.

        The center `Z(\Delta_q)` is generated by `\alpha`, `\beta`,
        `\gamma`, and `\Omega`.

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: AW.casimir_element()
            q*A*B*C + q^2*A^2 - q*A*a + (q^-2)*B^2 - (q^-1)*B*b + q^2*C^2 - q*C*g

        We check that the Casimir element is in the center::

            sage: Omega = AW.casimir_element()
            sage: all(Omega * gen == gen * Omega for gen in AW.algebra_generators())
            True
        """
        q = self._q
        I = self._indices
        d = {I((1, 1, 1, 0, 0, 0)): q,       # q ABC
             I((2, 0, 0, 0, 0, 0)): q**2,    # q^2 A^2
             I((0, 2, 0, 0, 0, 0)): q**-2,   # q^-2 B^2
             I((0, 0, 2, 0, 0, 0)): q**2,    # q^2 C^2
             I((1, 0, 0, 1, 0, 0)): -q,      # -q A\alpha
             I((0, 1, 0, 0, 1, 0)): -q**-1,  # -q^-1 B\beta
             I((0, 0, 1, 0, 0, 1)): -q}      # -q C\gamma
        return self.element_class(self, d)

    @cached_method
    def product_on_basis(self, x, y):
        """
        Return the product of the basis elements indexed by ``x`` and ``y``.

        INPUT:

        - ``x``, ``y`` -- tuple of length 6

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: AW.product_on_basis((0,0,0,0,0,0), (3,5,2,0,12,3))
            A^3*B^5*C^2*b^12*g^3
            sage: AW.product_on_basis((0,0,0,5,3,5), (3,5,2,0,12,3))
            A^3*B^5*C^2*a^5*b^15*g^8
            sage: AW.product_on_basis((7,0,0,5,3,5), (0,5,2,0,12,3))
            A^7*B^5*C^2*a^5*b^15*g^8
            sage: AW.product_on_basis((7,3,0,5,3,5), (0,2,2,0,12,3))
            A^7*B^5*C^2*a^5*b^15*g^8
            sage: AW.product_on_basis((0,1,0,5,3,5), (2,0,0,0,5,3))
            q^4*A^2*B*a^5*b^8*g^8 - (q^-3-q^5)*A*C*a^5*b^8*g^8
             + (1-q^4)*A*a^5*b^8*g^9 - (q^-4-2+q^4)*B*a^5*b^8*g^8
             + (q^-3-q^-1-q+q^3)*a^5*b^9*g^8
            sage: AW.product_on_basis((0,2,1,0,2,0), (1,1,0,2,1,0))
            q^4*A*B^3*C*a^2*b^3 - (q^5-q^9)*A^2*B^2*a^2*b^3
             + (q^2-q^4)*A*B^2*a^3*b^3 + (q^-3-q)*B^4*a^2*b^3
             - (q^-2-1)*B^3*a^2*b^4 - (q-q^9)*B^2*C^2*a^2*b^3
             + (1-q^4)*B^2*C*a^2*b^3*g + (q^-4+2-5*q^4+2*q^12)*A*B*C*a^2*b^3
             - (q^-1+q-2*q^3-2*q^5+q^7+q^9)*A*B*a^2*b^3*g
             - (q^-3-q^3-2*q^5+q^7+q^9)*B*C*a^3*b^3
             + (q^-2-1-q^2+q^4)*B*a^3*b^3*g
             - (q^-3-2*q+2*q^9-q^13)*A^2*a^2*b^3
             + (2*q^-2-2-3*q^2+3*q^4+q^10-q^12)*A*a^3*b^3
             + (q^-7-2*q^-3+2*q^5-q^9)*B^2*a^2*b^3
             - (q^-6-q^-4-q^-2+1-q^2+q^4+q^6-q^8)*B*a^2*b^4
             - (q^-7-q^-3-2*q+2*q^5+q^9-q^13)*C^2*a^2*b^3
             + (q^-6-3-2*q^2+5*q^4-q^8+q^10-q^12)*C*a^2*b^3*g
             - (q^-1-2*q+2*q^5-q^7)*a^4*b^3
             - (q^-3-q^-1-2*q+2*q^3+q^5-q^7)*a^2*b^3*g^2
        """
        I = self._indices
        # Commute the central parts to the right
        lhs = list(x[:3])
        rhs = list(y)
        for i in range(3, 6):
            rhs[i] += x[i]

        # No ABC variables on the RHS to move past
        if sum(rhs[:3]) == 0:
            return self.monomial(I(lhs + rhs[3:]))

        # We recurse using the PBW-type basis property:
        #   that YX = XY + lower order terms (see Theorem 4.1 in Terwilliger).
        q = self._q
        if lhs[2] > 0:  # lhs has a C
            if rhs[0] > 0:  # rhs has an A to commute with C
                lhs[2] -= 1
                rhs[0] -= 1
                rel = {I((1, 0, 1, 0, 0, 0)): q**-2,         # q^2 AC
                       I((0, 1, 0, 0, 0, 0)): q**-3 - q**1,  # q^-1(q^-2-q^2) B
                       I((0, 0, 0, 0, 1, 0)): 1 - q**-2}     # -q^-1(q^-1-q) b
                rel = self.element_class(self, rel)
                return self.monomial(I(lhs+[0]*3)) * (rel * self.monomial(I(rhs)))
            elif rhs[1] > 0:  # rhs has a B to commute with C
                lhs[2] -= 1
                rhs[1] -= 1
                rel = {I((0, 1, 1, 0, 0, 0)): q**2,           # q^2 BC
                       I((1, 0, 0, 0, 0, 0)): q**3 - q**-1,  # q(q^2-q^-2) A
                       I((0, 0, 0, 1, 0, 0)): -q**2 + 1}     # -q(q-q^-1) a
                rel = self.element_class(self, rel)
                return self.monomial(I(lhs+[0]*3)) * (rel * self.monomial(I(rhs)))
            else:  # nothing to commute as rhs has no A nor B
                rhs[2] += lhs[2]
                rhs[1] = lhs[1]
                rhs[0] = lhs[0]
                return self.monomial(I(rhs))

        elif lhs[1] > 0:  # lhs has a B
            if rhs[0] > 0:  # rhs has an A to commute with B
                lhs[1] -= 1
                rhs[0] -= 1
                rel = {I((1, 1, 0, 0, 0, 0)): q**2,          # q^2 AB
                       I((0, 0, 1, 0, 0, 0)): q**3 - q**-1,  # q(q^2-q^-2) C
                       I((0, 0, 0, 0, 0, 1)): -q**2 + 1}     # -q(q-q^-1) g
                rel = self.element_class(self, rel)
                return self.monomial(I(lhs+[0]*3)) * (rel * self.monomial(I(rhs)))
            else:  # nothing to commute as rhs has no A
                rhs[1] += lhs[1]
                rhs[0] = lhs[0]
                return self.monomial(I(rhs))

        elif lhs[0] > 0:  # lhs has an A
            rhs[0] += lhs[0]
            return self.monomial(I(rhs))

        # otherwise, lhs is just 1
        return self.monomial(I(rhs))

    def permutation_automorphism(self):
        r"""
        Return the permutation automorphism `\rho` of ``self``.

        We define the automorphism `\rho` by

        .. MATH::

            A \mapsto B \mapsto C \mapsto A,
            \qquad
            \alpha \mapsto \beta \mapsto \gamma \mapsto \alpha.

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: rho = AW.permutation_automorphism()
            sage: [rho(gen) for gen in AW.algebra_generators()]
            [B, C, A, b, g, a]

            sage: AW.an_element()
            (q^-3+3+2*q+q^2)*a*b*g^3 + q*A*C^2*b + 3*q^2*B*a^2*g + A
            sage: rho(AW.an_element())
            (q^-3+3+2*q+q^2)*a^3*b*g + q^5*A^2*B*g + 3*q^2*C*a*b^2
             - (q^-2-q^6)*A*C*g + (q-q^5)*A*g^2 - (q^-3-2*q+q^5)*B*g
             + (q^-2-1-q^2+q^4)*b*g + B

            sage: r3 = rho * rho * rho
            sage: [r3(gen) for gen in AW.algebra_generators()]
            [A, B, C, a, b, g]
            sage: r3(AW.an_element()) == AW.an_element()
            True
        """
        A,B,C,a,b,g = self.gens()
        return AlgebraMorphism(self, [B,C,A,b,g,a], codomain=self)

    rho = permutation_automorphism

    def reflection_automorphism(self):
        r"""
        Return the reflection automorphism `\sigma` of ``self``.

        We define the automorphism `\sigma` by

        .. MATH::

            A \mapsto B \mapsto A,
            \qquad
            C \mapsto C + \frac{AB - BA}{q-q^{-1}}
            = C - qAB - (1+q^2) C + q \gamma,

        .. MATH::

            \alpha \mapsto \beta \mapsto \alpha,
            \gamma \mapsto \gamma.

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: sigma = AW.reflection_automorphism()
            sage: [sigma(gen) for gen in AW.algebra_generators()]
            [B, A, -q*A*B - q^2*C + q*g, b, a, g]

            sage: AW.an_element()
            (q^-3+3+2*q+q^2)*a*b*g^3 + q*A*C^2*b + 3*q^2*B*a^2*g + A
            sage: sigma(AW.an_element())
            q^9*A^2*B^3*a + (q^10+q^14)*A*B^2*C*a - (q^7+q^9)*A*B^2*a*g
             + (q^-3+3+2*q+q^2)*a*b*g^3 + (q-3*q^9+q^13+q^17)*A^2*B*a
             - (q^2-q^6-q^8+q^14)*A*B*a^2 + 3*q^2*A*b^2*g + (q^5-q^9)*B^3*a
             - (q^6-q^8)*B^2*a*b + q^13*B*C^2*a - 2*q^10*B*C*a*g + q^7*B*a*g^2
             + (q^2-2*q^10+q^18)*A*C*a - (q-q^7-2*q^9+2*q^11-q^15+q^17)*A*a*g
             - (q^3-q^7-q^9+q^13)*C*a^2 + (q^2-q^6-2*q^8+2*q^10)*a^2*g
             + (q-3*q^5+3*q^9-q^13)*B*a - (q^2-q^4-2*q^6+2*q^8+q^10-q^12)*a*b + B

            sage: s2 = sigma * sigma
            sage: [s2(gen) for gen in AW.algebra_generators()]
            [A, B, C, a, b, g]
            sage: s2(AW.an_element()) == AW.an_element()
            True
        """
        A,B,C,a,b,g = self.gens()
        q = self._q
        # Note that sage: (A*B-B*A) / (q-q^-1) == -q*A*B - (1+q^2)*C + q*g
        Cp = C - q*A*B - (1+q**2)*C + q*g
        return AlgebraMorphism(self, [B,A,Cp,b,a,g], codomain=self)

    sigma = reflection_automorphism

    def loop_representation(self):
        r"""
        Return the map `\pi` from ``self`` to `2 \times 2` matrices
        over `R[\lambda,\lambda^{-1}]`, where `F` is the fraction field
        of the base ring of ``self``.

        Let `AW` be the Askey-Wilson algebra over `R`, and let `F` be
        the fraction field of `R`. Let `M` be the space of `2 \times 2`
        matrices over `F[\lambda, \lambda^{-1}]`. Consider the following
        elements of `M`:

        .. MATH::

            \mathcal{A} = \begin{pmatrix}
                \lambda & 1 - \lambda^{-1} \\ 0 & \lambda^{-1}
            \end{pmatrix},
            \qquad
            \mathcal{B} = \begin{pmatrix}
                \lambda^{-1} & 0 \\ \lambda - 1 & \lambda
            \end{pmatrix},
            \qquad
            \mathcal{C} = \begin{pmatrix}
                1 & \lambda - 1 \\ 1 - \lambda^{-1} & \lambda + \lambda^{-1} - 1
            \end{pmatrix}.

        From Lemma 3.11 of [Terwilliger2011]_, we define a
        representation `\pi: AW \to M` by

        .. MATH::

            A \mapsto q \mathcal{A} + q^{-1} \mathcal{A}^{-1},
            \qquad
            B \mapsto q \mathcal{B} + q^{-1} \mathcal{B}^{-1},
            \qquad
            C \mapsto q \mathcal{C} + q^{-1} \mathcal{C}^{-1},

        .. MATH::

            \alpha, \beta, \gamma \mapsto \nu I,

        where `\nu = (q^2 + q^-2)(\lambda + \lambda^{-1})
        + (\lambda + \lambda^{-1})^2`.

        We call this representation the *loop representation* as
        it is a representation using the loop group
        `SL_2(F[\lambda,\lambda^{-1}])`.

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: q = AW.q()
            sage: pi = AW.loop_representation()
            sage: A,B,C,a,b,g = [pi(gen) for gen in AW.algebra_generators()]
            sage: A
            [                1/q*lambda^-1 + q*lambda ((-q^2 + 1)/q)*lambda^-1 + ((q^2 - 1)/q)]
            [                                       0                 q*lambda^-1 + 1/q*lambda]
            sage: B
            [             q*lambda^-1 + 1/q*lambda                                     0]
            [((-q^2 + 1)/q) + ((q^2 - 1)/q)*lambda              1/q*lambda^-1 + q*lambda]
            sage: C
            [1/q*lambda^-1 + ((q^2 - 1)/q) + 1/q*lambda      ((q^2 - 1)/q) + ((-q^2 + 1)/q)*lambda]
            [  ((q^2 - 1)/q)*lambda^-1 + ((-q^2 + 1)/q)    q*lambda^-1 + ((-q^2 + 1)/q) + q*lambda]
            sage: a
            [lambda^-2 + ((q^4 + 1)/q^2)*lambda^-1 + 2 + ((q^4 + 1)/q^2)*lambda + lambda^2                                                                             0]
            [                                                                            0 lambda^-2 + ((q^4 + 1)/q^2)*lambda^-1 + 2 + ((q^4 + 1)/q^2)*lambda + lambda^2]
            sage: a == b
            True
            sage: a == g
            True

            sage: AW.an_element()
            (q^-3+3+2*q+q^2)*a*b*g^3 + q*A*C^2*b + 3*q^2*B*a^2*g + A
            sage: x = pi(AW.an_element())
            sage: y = (q^-3+3+2*q+q^2)*a*b*g^3 + q*A*C^2*b + 3*q^2*B*a^2*g + A
            sage: x == y
            True

        We check the defining relations of the Askey-Wilson algebra::

            sage: A + (q*B*C - q^-1*C*B) / (q^2 - q^-2) == a / (q + q^-1)
            True
            sage: B + (q*C*A - q^-1*A*C) / (q^2 - q^-2) == b / (q + q^-1)
            True
            sage: C + (q*A*B - q^-1*B*A) / (q^2 - q^-2) == g / (q + q^-1)
            True

        We check Lemma 3.12 in [Terwilliger2011]_::

            sage: M = pi.codomain()
            sage: la = M.base_ring().gen()
            sage: p = M([[0,-1],[1,1]])
            sage: s = M([[0,1],[la,0]])
            sage: rho = AW.rho()
            sage: sigma = AW.sigma()
            sage: all(p*pi(gen)*~p == pi(rho(gen)) for gen in AW.algebra_generators())
            True
            sage: all(s*pi(gen)*~s == pi(sigma(gen)) for gen in AW.algebra_generators())
            True
        """
        from sage.matrix.matrix_space import MatrixSpace
        q = self._q
        base = LaurentPolynomialRing(self.base_ring().fraction_field(), 'lambda')
        la = base.gen()
        inv = ~la
        M = MatrixSpace(base, 2)
        A = M([[la,1-inv],[0,inv]])
        Ai = M([[inv,inv-1],[0,la]])
        B = M([[inv,0],[la-1,la]])
        Bi = M([[la,0],[1-la,inv]])
        C = M([[1,1-la],[inv-1,la+inv-1]])
        Ci = M([[la+inv-1,la-1],[1-inv,1]])
        mu = la + inv
        nu = (self._q**2 + self._q**-2) * mu + mu**2
        nuI = M(nu)
        # After #29374 is fixed, the category can become
        # Algebras(Rings().Commutative()) as it was before #29399.
        category = Rings()
        return AlgebraMorphism(self, [q*A + q**-1*Ai, q*B + q**-1*Bi, q*C + q**-1*Ci,
                                      nuI, nuI, nuI],
                               codomain=M, category=category)

    pi = loop_representation


def _basis_key(t):
    """
    Return a key for the basis element of the Askey-Wilson algebra
    indexed by ``t``.

    EXAMPLES::

        sage: from sage.algebras.askey_wilson import _basis_key
        sage: I = algebras.AskeyWilson(QQ).indices()
        sage: _basis_key(I((0,2,3,1,2,5)))
        (13, (0, 2, 3, 1, 2, 5))
    """
    return (sum(t), t.value)


class AlgebraMorphism(ModuleMorphismByLinearity):
    """
    An algebra morphism of the Askey-Wilson algebra defined by
    the images of the generators.
    """
    def __init__(self, domain, on_generators, position=0, codomain=None,
                 category=None):
        """
        Given a map on the multiplicative basis of a free algebra, this method
        returns the algebra morphism that is the linear extension of its image
        on generators.

        INPUT:

        - ``domain`` -- an Askey-Wilson algebra
        - ``on_generators`` -- list of length 6 corresponding to
          the images of the generators
        - ``codomain`` -- (optional) the codomain
        - ``position`` -- integer (default: 0)
        - ``category`` -- (optional) category

        OUTPUT: module morphism

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: sigma = AW.sigma()
            sage: TestSuite(sigma).run()
        """
        if category is None:
            category = Algebras(Rings().Commutative()).WithBasis()
        self._on_generators = tuple(on_generators)
        ModuleMorphismByLinearity.__init__(self, domain=domain, codomain=codomain,
                                           position=position, category=category)

    def __eq__(self, other):
        """
        Check equality.

        EXAMPLES::

            sage: from sage.algebras.askey_wilson import AlgebraMorphism
            sage: AW = algebras.AskeyWilson(QQ)
            sage: rho = AW.rho()
            sage: sigma = AW.sigma()
            sage: id = AlgebraMorphism(AW, AW.gens(), codomain=AW)
            sage: sigma * sigma == id
            True
            sage: id == rho * rho * rho
            True
        """
        return (self.__class__ is other.__class__ and self.parent() == other.parent()
                and self._zero == other._zero
                and self._on_generators == other._on_generators
                and self._position == other._position
                and self._is_module_with_basis_over_same_base_ring
                    == other._is_module_with_basis_over_same_base_ring)

    def _on_basis(self, c):
        r"""
        Compute the image of this morphism on the basis element
        indexed by ``c``.

        INPUT:

        - ``c`` -- tuple of length 6

        OUTPUT: element of the codomain

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: AW.inject_variables()
            Defining A, B, C, a, b, g
            sage: rho = AW.rho()
            sage: sigma = AW.sigma()
            sage: rho._on_basis((2,1,1,2,0,5))
            q^2*A*B^2*C*a^5*b^2 + (q^-3-q)*B^3*a^5*b^2 - (q^-2-1)*B^2*a^5*b^3
             - (q^-3-q^5)*B*C^2*a^5*b^2 + (q^-2-q^2)*B*C*a^5*b^2*g
             + (q^-2-2*q^2+q^6)*A*C*a^5*b^2 - (q^-1-q-q^3+q^5)*C*a^6*b^2

            sage: sigma._on_basis((2,1,1,2,0,5))
            -q^9*A^2*B^3*b^2*g^5 + (q^2-q^10-q^14)*A*B^2*C*b^2*g^5
             - (q^3-q^7-q^9)*A*B^2*b^2*g^6
             - (2*q-q^5-3*q^9+q^13+q^17)*A^2*B*b^2*g^5
             + (1-q^6-q^8+q^14)*A*B*a*b^2*g^5 + (q^-3-q-q^5+q^9)*B^3*b^2*g^5
             - (q^-2-1-q^6+q^8)*B^2*b^3*g^5 + (q^5-q^13)*B*C^2*b^2*g^5
             - (q^2+q^6-2*q^10)*B*C*b^2*g^6 + (q^3-q^7)*B*b^2*g^7
             + (q^-6-4*q^2+2*q^6+2*q^10-q^18)*A*C*b^2*g^5
             - (q^-3+q^-1-3*q-2*q^3+q^5+2*q^7+2*q^9-2*q^11+q^15-q^17)*A*b^2*g^6
             - (q^-3-2*q-q^3+q^5+q^7+q^9-q^13)*C*a*b^2*g^5
             + (q^-2-1-q^2-q^4+2*q^6+2*q^8-2*q^10)*a*b^2*g^6
             + (q^-7-3*q^-3+2*q+2*q^5-3*q^9+q^13)*B*b^2*g^5
             - (q^-6-q^-4-2*q^-2+2+2*q^6-2*q^8-q^10+q^12)*b^3*g^5

            sage: rho(B*A)
            q^2*B*C - (q^-1-q^3)*A + (1-q^2)*a
            sage: rho(A*B)
            B*C
            sage: rho(A*B*C)
            A*B*C + (q^-3-q)*B^2 - (q^-2-1)*B*b - (q^-3-q)*C^2 + (q^-2-1)*C*g
            sage: rho(B*C*A)
            A*B*C - (q^-3-q)*A^2 + (q^-2-1)*A*a + (q^-3-q)*B^2 - (q^-2-1)*B*b
            sage: rho(C*A*B)
            A*B*C

            sage: rho(C^2*a*b^6*g^2)
            A^2*a^2*b*g^6

            sage: sigma(C^2)
            q^4*A^2*B^2 + (q^3+q^7)*A*B*C - (q^2+q^4)*A*B*g
             - (q^4-q^8)*A^2 + (q^5-q^7)*A*a + (1-q^4)*B^2
             - (q-q^3)*B*b + q^4*C^2 - 2*q^3*C*g + q^2*g^2
            sage: sigma(A*B)
            q^2*A*B - (q^-1-q^3)*C + (1-q^2)*g
            sage: sigma(C + 3*g*A*B)
            3*q^2*A*B*g - q*A*B - (3*q^-1-3*q^3)*C*g
             + (3-3*q^2)*g^2 - q^2*C + q*g
        """
        return self.codomain().prod(self._on_generators[i]**exp
                                    for i, exp in enumerate(c))

    def _composition_(self, right, homset):
        """
        Return the composition of ``self`` and ``right`` in ``homset``.

        EXAMPLES::

            sage: AW = algebras.AskeyWilson(QQ)
            sage: rho = AW.rho()
            sage: sigma = AW.sigma()
            sage: s2 = sigma * sigma
            sage: s2._on_generators
            (A, B, C, a, b, g)
            sage: sr = sigma * rho
            sage: sr._on_generators
            (C, B, -q*B*C - q^2*A + q*a, g, b, a)
            sage: rs = rho * sigma
            sage: rs._on_generators
            (A, -q*A*B - q^2*C + q*g, B, a, g, b)
        """
        if isinstance(right, AlgebraMorphism):
            cat = homset.homset_category()
            return AlgebraMorphism(homset.domain(),
                                   [right(g) for g in self._on_generators],
                                   codomain=homset.codomain(), category=cat)
        return super()._composition_(right, homset)



# sage_setup: distribution = sagemath-combinat
cdef extern from 'symmetrica/def.h':
    INT plethysm(OP s1, OP s2, OP res)
    INT schur_schur_plet(OP p1, OP p2, OP res)


def plethysm_symmetrica(outer, inner):
    cdef OP couter = callocobject(), cinner = callocobject(), cresult = callocobject()

    _op_schur(outer, couter)
    _op_schur(inner, cinner)

    sig_on()
    plethysm(couter, cinner, cresult)
    sig_off()

    res = _py(cresult)

    freeall(couter)
    freeall(cinner)
    freeall(cresult)

    return res


def schur_schur_plet_symmetrica(outer, inner):
    cdef OP couter = callocobject(), cinner = callocobject(), cresult = callocobject()

    _op_partition(outer, couter)
    _op_partition(inner, cinner)

    sig_on()
    schur_schur_plet(couter, cinner, cresult)
    sig_off()

    res = _py(cresult)

    freeall(couter)
    freeall(cinner)
    freeall(cresult)

    return res



# sage_setup: distribution = sagemath-combinat
r"""
Super partitions

A super partition of size `n` and fermionic sector `m` is a
pair consisting of a strict partition of some integer `r` of
length `m` (that may end in a `0`) and an integer partition of
`n - r`.

This module provides tools for manipulating super partitions.

Super partitions are the indexing set for symmetric functions in
super space.

Super partitions may be input in two different formats: one as a pair
consisting of fermionic (strict partition) and a bosonic (partition) part
and the other as a list of integer values where the negative entries come
first and are listed in strict order followed by the positive values in
weak order.

A super partition is displayed as two partitions separated by a semicolon
as a default.  Super partitions may also be displayed as a weakly increasing
sequence of integers that are strict if the numbers are not positive.

These combinatorial objects index the space of symmetric polynomials in
two sets of variables, one commuting and one anti-commuting, and they
are known as symmetric functions in super space (hence the origin of the
name super partitions).

EXAMPLES::

    sage: SuperPartitions()
    Super Partitions
    sage: SuperPartitions(2)
    Super Partitions of 2
    sage: SuperPartitions(2).cardinality()
    8
    sage: SuperPartitions(4,2)
    Super Partitions of 4 and of fermionic sector 2
    sage: [[2,0],[1,1]] in SuperPartitions(4,2)
    True
    sage: [[1,0],[1,1]] in SuperPartitions(4,2)
    False
    sage: [[1,0],[2,1]] in SuperPartitions(4)
    True
    sage: [[1,0],[2,2,1]] in SuperPartitions(4)
    False
    sage: [[1,0],[2,1]] in SuperPartitions()
    True
    sage: [[1,1],[2,1]] in SuperPartitions()
    False
    sage: [-2, 0, 1, 1] in SuperPartitions(4,2)
    True
    sage: [-1, 0, 1, 1] in SuperPartitions(4,2)
    False
    sage: [-2, -2, 2, 1] in SuperPartitions(7,2)
    False

REFERENCES:

- [JL2016]_

AUTHORS:

- Mike Zabrocki
"""

# ****************************************************************************
#       Copyright (C) 2018 Mike Zabrocki <zabrocki at mathstat.yorku.ca>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from __future__ import annotations
from functools import reduce
from itertools import chain

from sage.structure.list_clone import ClonableArray
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.parent import Parent
from sage.structure.richcmp import richcmp, richcmp_method
from sage.combinat.partition import Partition, Partitions, _Partitions
from sage.combinat.composition import Composition
from sage.categories.enumerated_sets import EnumeratedSets
from sage.rings.integer import Integer
from sage.structure.global_options import GlobalOptions
from sage.rings.integer_ring import ZZ
from sage.misc.inherit_comparison import InheritComparisonClasscallMetaclass


@richcmp_method
class SuperPartition(ClonableArray,
                     metaclass=InheritComparisonClasscallMetaclass):
    r"""
    A super partition.

    A *super partition* of size `n` and fermionic sector `m` is a
    pair consisting of a strict partition of some integer `r` of
    length `m` (that may end in a `0`) and an integer partition of
    `n - r`.

    EXAMPLES::

        sage: sp = SuperPartition([[1,0],[2,2,1]]); sp
        [1, 0; 2, 2, 1]
        sage: sp[0]
        (1, 0)
        sage: sp[1]
        (2, 2, 1)
        sage: sp.fermionic_degree()
        2
        sage: sp.bosonic_degree()
        6
        sage: sp.length()
        5
        sage: sp.conjugate()
        [4, 2; ]
    """
    @staticmethod
    def __classcall_private__(cls, lst):
        r"""
        Construct a superpartition in the correct parent.

        EXAMPLES::

            sage: SuperPartition([[1],[1]]).parent()
            Super Partitions
            sage: SuperPartition([[1],[1]])
            [1; 1]
            sage: SuperPartition([-1, 1])
            [1; 1]
            sage: SuperPartition([[1,1],[1]])
            Traceback (most recent call last):
            ...
            ValueError: [[1, 1], [1]] not in Super Partitions
            sage: SuperPartition([-1,1])
            [1; 1]
            sage: SuperPartition([])
            [; ]

            sage: SP = SuperPartitions(8,4)([[3,2,1,0],[2]])
            sage: SuperPartition(SP) is SP
            True
        """
        if isinstance(lst, SuperPartition):
            return lst
        SPs = SuperPartitions()
        if not lst:
            return SPs([[], []])
        elif isinstance(lst[0], (list, tuple)):
            return SPs([[Integer(a) for a in lst[0]],
                        [Integer(a) for a in lst[1]]])
        else:
            return SPs([[-a for a in lst if a <= 0],
                        [a for a in lst if a > 0]])

    def __init__(self, parent, lst, check=True, immutable=True):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: SP = SuperPartition([[1],[1]])
            sage: TestSuite(SP).run()
        """
        if check and lst not in parent:
            raise ValueError("%s not in %s" % (lst, parent))
        lst = [tuple(lst[0]), tuple(lst[1])]
        ClonableArray.__init__(self, parent, lst, False, immutable)

    def check(self):
        """
        Check that ``self`` is a valid super partition.

        EXAMPLES::

            sage: SP = SuperPartition([[1],[1]])
            sage: SP.check()
        """
        if self not in self.parent():
            raise ValueError("%s not in %s" % (self, self.parent()))

    def __richcmp__(self, other, op) -> bool:
        r"""
        Check whether ``self`` is equal to ``other``.

        .. TODO::

            This overwrites the equality check of
            :class:`~sage.structure.list_clone.ClonableArray`
            in order to circumvent the coercion framework.
            Eventually this should be solved more elegantly.

            For now, two elements are compared by their defining lists.
        """
        if isinstance(other, SuperPartition):
            return richcmp(list(self), list(other), op)
        else:
            return richcmp(list(self), other, op)

    def _repr_(self) -> str:
        r"""
        Return a string representation of ``self``.

        A super partition is represented by the antisymmetric and symmetric
        parts separated by a semicolon.

        EXAMPLES::

            sage: SuperPartition([[1],[1]])
            [1; 1]
            sage: SuperPartition([[],[1]])
            [; 1]
            sage: SuperPartition([])
            [; ]
            sage: SuperPartitions.options.display = "list"
            sage: SuperPartition([[1],[1]])
            [-1, 1]
            sage: SuperPartition([[],[1]])
            [1]
            sage: SuperPartition([-2,-1,0,2,1])
            [-2, -1, 0, 2, 1]
            sage: SuperPartitions.options.display = "pair"
            sage: SuperPartition([[1],[1]])
            [[1], [1]]
            sage: SuperPartition([[],[1]])
            [[], [1]]
            sage: SuperPartition([-2,-1,0,2,1])
            [[2, 1, 0], [2, 1]]
            sage: SuperPartitions.options._reset()
        """
        display = self.parent().options.display
        if display == "default":
            asp = ', '.join(str(a) for a in self.antisymmetric_part())
            sp = ', '.join(str(a) for a in self.symmetric_part())
            return '[' + asp + '; ' + sp + ']'
        elif display == "pair":
            return self._repr_pair()
        else:  # "list"
            return self._repr_list()

    def _repr_pair(self) -> str:
        r"""
        Representation of a super partition as a pair.

        A super partition is represented by a list consisting of the
        antisymmetric and symmetric parts.

        EXAMPLES::

            sage: SuperPartition([[1],[1]])._repr_pair()
            '[[1], [1]]'
            sage: SuperPartition([[],[1]])._repr_pair()
            '[[], [1]]'
            sage: SuperPartition([[],[]])._repr_pair()
            '[[], []]'
        """
        return repr(self.to_list())

    def _repr_list(self) -> str:
        r"""
        Representation of a super partition as a list.

        A super partition is represented by a list consisting of the
        negative values for the antisymmetric part listed first followed
        by positive values for the symmetric part

        EXAMPLES::

            sage: SuperPartition([[1],[1]])._repr_list()
            '[-1, 1]'
            sage: SuperPartition([[],[1]])._repr_list()
            '[1]'
            sage: SuperPartition([[],[]])._repr_list()
            '[]'
        """
        return repr([-a for a in self[0]] + list(self[1]))

    def _latex_(self) -> str:
        r"""
        Latex a super partition.

        A super partition is represented by the antisymmetric and symmetric
        parts separated by a semicolon.

        EXAMPLES::

            sage: latex(SuperPartition([[1],[1]]))
            (1; 1)
            sage: latex(SuperPartition([[],[1]]))
            (; 1)
        """
        return ('(' + ','.join(str(a) for a in self.antisymmetric_part())
                + '; ' + ', '.join(str(a) for a in self.symmetric_part()) + ')')

    def to_list(self) -> list:
        r"""
        The list of two lists with the antisymmetric and symmetric parts.

        EXAMPLES::

            sage: SuperPartition([[1],[1]]).to_list()
            [[1], [1]]
            sage: SuperPartition([[],[1]]).to_list()
            [[], [1]]
        """
        return [list(self[0]), list(self[1])]

    def to_composition(self) -> Composition:
        r"""
        Concatenate the antisymmetric and symmetric parts to a composition.

        OUTPUT:

        - a (possibly weak) composition

        EXAMPLES::

            sage: SuperPartition([[3,1],[2,2,1]]).to_composition()
            [3, 1, 2, 2, 1]
            sage: SuperPartition([[2,1,0],[3,3]]).to_composition()
            [2, 1, 0, 3, 3]
            sage: SuperPartition([[2,1,0],[3,3]]).to_composition().parent()
            Compositions of nonnegative integers
        """
        return Composition(self[0] + self[1])

    def to_partition(self) -> Partition:
        r"""
        Concatenate and sort the antisymmetric and symmetric parts
        to a partition.

        OUTPUT: a partition

        EXAMPLES::

            sage: SuperPartition([[3,1],[2,2,1]]).to_partition()
            [3, 2, 2, 1, 1]
            sage: SuperPartition([[2,1,0],[3,3]]).to_partition()
            [3, 3, 2, 1]
            sage: SuperPartition([[2,1,0],[3,3]]).to_partition().parent()
            Partitions
        """
        return Partition(sorted(self[0] + self[1], reverse=True))  # type:ignore

    def antisymmetric_part(self) -> list:
        r"""
        The antisymmetric part as a list of strictly decreasing integers.

        OUTPUT: list

        EXAMPLES::

            sage: SuperPartition([[3,1],[2,2,1]]).antisymmetric_part()
            [3, 1]
            sage: SuperPartition([[2,1,0],[3,3]]).antisymmetric_part()
            [2, 1, 0]
        """
        return list(self[0])

    a_part = antisymmetric_part

    def symmetric_part(self) -> list:
        r"""
        The symmetric part as a list of weakly decreasing integers.

        OUTPUT: list

        EXAMPLES::

            sage: SuperPartition([[3,1],[2,2,1]]).symmetric_part()
            [2, 2, 1]
            sage: SuperPartition([[2,1,0],[3,3]]).symmetric_part()
            [3, 3]
        """
        return list(self[1])

    s_part = symmetric_part

    def bosonic_degree(self) -> int:
        r"""
        Return the bosonic degree of ``self``.

        The *bosonic degree* is the sum of the sizes of the
        antisymmetric and symmetric parts.

        OUTPUT: integer

        EXAMPLES::

            sage: SuperPartition([[3,1],[2,2,1]]).bosonic_degree()
            9
            sage: SuperPartition([[2,1,0],[3,3]]).bosonic_degree()
            9
        """
        return sum(self.antisymmetric_part() + self.symmetric_part())

    degree = bosonic_degree

    def fermionic_degree(self) -> int:
        r"""
        Return the fermionic degree of ``self``.

        The *fermionic degree* is the length of the antisymmetric part.

        OUTPUT: integer

        EXAMPLES::

            sage: SuperPartition([[3,1],[2,2,1]]).fermionic_degree()
            2
            sage: SuperPartition([[2,1,0],[3,3]]).fermionic_degree()
            3
        """
        return len(self.antisymmetric_part())

    fermionic_sector = fermionic_degree

    def bi_degree(self) -> tuple:
        r"""
        Return the bidegree of ``self``, which is a pair consisting
        of the bosonic and fermionic degree.

        OUTPUT: a tuple of two integers

        EXAMPLES::

            sage: SuperPartition([[3,1],[2,2,1]]).bi_degree()
            (9, 2)
            sage: SuperPartition([[2,1,0],[3,3]]).bi_degree()
            (9, 3)
        """
        return (self.bosonic_degree(), self.fermionic_degree())

    def length(self) -> int:
        r"""
        Return the length of ``self``, which is the sum of the
        lengths of the antisymmetric and symmetric part.

        OUTPUT: integer

        EXAMPLES::

            sage: SuperPartition([[3,1],[2,2,1]]).length()
            5
            sage: SuperPartition([[2,1,0],[3,3]]).length()
            5
        """
        return self.fermionic_degree() + len(self.symmetric_part())

    def bosonic_length(self) -> int:
        r"""
        Return the length of the partition of the symmetric part.

        OUTPUT: integer

        EXAMPLES::

            sage: SuperPartition([[3,1],[2,2,1]]).bosonic_length()
            3
            sage: SuperPartition([[2,1,0],[3,3]]).bosonic_length()
            2
        """
        return len(self.symmetric_part())

    def shape_circled_diagram(self) -> Partition:
        r"""
        A concatenated partition with an extra cell for each antisymmetric part

        OUTPUT: a partition

        EXAMPLES::

            sage: SuperPartition([[3,1],[2,2,1]]).shape_circled_diagram()
            [4, 2, 2, 2, 1]
            sage: SuperPartition([[2,1,0],[3,3]]).shape_circled_diagram()
            [3, 3, 3, 2, 1]
        """
        pi = sorted([a + 1 for a in self.antisymmetric_part()] +
                    self.symmetric_part(), reverse=True)
        return Partition(pi)  # type:ignore

    @staticmethod
    def from_circled_diagram(shape, corners) -> SuperPartition:
        r"""
        Construct a super partition from a circled diagram.

        A circled diagram consists of a partition of the concatenation of
        the antisymmetric and symmetric parts and a list of addable cells
        of the partition which indicate the location of the circled cells.

        INPUT:

        - ``shape`` -- a partition or list of integers
        - ``corners`` -- list of removable cells of ``shape``

        OUTPUT: a :class:`SuperPartition`

        EXAMPLES::

            sage: SuperPartition.from_circled_diagram([3, 2, 2, 1, 1], [(0, 3), (3, 1)])
            [3, 1; 2, 2, 1]
            sage: SuperPartition.from_circled_diagram([3, 3, 2, 1], [(2, 2), (3, 1), (4, 0)])
            [2, 1, 0; 3, 3]
            sage: from_cd = SuperPartition.from_circled_diagram
            sage: all(sp == from_cd(*sp.to_circled_diagram()) for sp in SuperPartitions(4))
            True
        """
        data = [sorted([c[1] for c in corners], reverse=True),
                [shape[i] for i in range(len(shape))
                 if i not in [c[0] for c in corners]]]
        return SuperPartition(data)  # type:ignore

    def to_circled_diagram(self) -> list:
        r"""
        The shape of the circled diagram and a list of addable cells

        A circled diagram consists of a partition for the outer shape
        and a list of removable cells of the partition indicating the
        location of the circled cells

        OUTPUT: list consisting of a partition and a list of pairs of integers

        EXAMPLES::

            sage: SuperPartition([[3,1],[2,2,1]]).to_circled_diagram()
            [[3, 2, 2, 1, 1], [(0, 3), (3, 1)]]
            sage: SuperPartition([[2,1,0],[3,3]]).to_circled_diagram()
            [[3, 3, 2, 1], [(2, 2), (3, 1), (4, 0)]]
            sage: from_cd = SuperPartition.from_circled_diagram
            sage: all(sp == from_cd(*sp.to_circled_diagram()) for sp in SuperPartitions(4))
            True
        """
        shape = self.to_partition()
        corners = [c for c in shape.addable_cells() if c[1] in self.antisymmetric_part()]
        return [shape, corners]

    def conjugate(self) -> SuperPartition:
        r"""
        Conjugate of a super partition.

        The *conjugate* of a super partition is defined by conjugating
        the circled diagram.

        OUTPUT: a :class:`SuperPartition`

        EXAMPLES::

            sage: SuperPartition([[3, 1, 0], [4, 3, 2, 1]]).conjugate()
            [6, 4, 1; 3]
            sage: all(sp == sp.conjugate().conjugate() for sp in SuperPartitions(4))
            True
            sage: all(sp.conjugate() in SuperPartitions(3,2) for sp in SuperPartitions(3,2))
            True
        """
        sd = self.to_circled_diagram()
        return SuperPartition.from_circled_diagram(sd[0].conjugate(),
                                                   [(j, i) for (i, j) in sd[1]])

    def zee(self) -> Integer:
        r"""
        Return the centralizer size of a permutation of cycle
        type symmetric part of ``self``.

        OUTPUT: a positive integer

        EXAMPLES::

            sage: SuperPartition([[1,0],[3,1,1]]).zee()
            6
            sage: SuperPartition([[1],[2,2,1]]).zee()
            8
            sage: sum(1/sp.zee() for sp in SuperPartitions(6,0))
            1
        """
        return Partition(self.symmetric_part()).centralizer_size()  # type:ignore

    def sign(self) -> int:
        r"""
        Return the sign of a permutation of cycle type the
        symmetric part of ``self``.

        OUTPUT: either `1` or `-1`

        EXAMPLES::

            sage: SuperPartition([[1,0],[3,1,1]]).sign()
            -1
            sage: SuperPartition([[1,0],[3,2,1]]).sign()
            1
            sage: sum(sp.sign()/sp.zee() for sp in SuperPartitions(6,0))
            0
        """
        return (-1)**(self.degree() - len(self.symmetric_part()))

    def dominates(self, other) -> bool:
        r"""
        Return ``True`` if and only if ``self`` dominates ``other``.

        If the symmetric and anti-symmetric parts of ``self`` and ``other``
        are not the same size then the result is ``False``.

        EXAMPLES::

            sage: LA = SuperPartition([[2,1],[2,1,1]])
            sage: LA.dominates([[2,1],[3,1]])
            False
            sage: LA.dominates([[2,1],[1,1,1,1]])
            True
            sage: LA.dominates([[3],[2,1,1]])
            False
            sage: LA.dominates([[1],[1]*6])
            False
        """
        return (self.degree() == sum(other[0]) + sum(other[1]) and
                Partition(self.antisymmetric_part()).dominates(other[0]) and
                Partition(self.symmetric_part()).dominates(other[1]))

    def add_horizontal_border_strip_star(self, h) -> list:
        r"""
        Return a list of super partitions that differ from ``self``
        by a horizontal strip.

        The notion of horizontal strip comes from the Pieri rule for the
        Schur-star basis of symmetric functions in super space (see
        Theorem 7 from [JL2016]_).

        INPUT:

        - ``h`` -- number of cells in the horizontal strip

        OUTPUT: list of super partitions

        EXAMPLES::

            sage: SuperPartition([[4,1],[3]]).add_horizontal_border_strip_star(3)
            [[3, 1; 7],
             [4, 1; 6],
             [3, 0; 6, 2],
             [3, 1; 6, 1],
             [4, 0; 5, 2],
             [4, 1; 5, 1],
             [3, 0; 5, 3],
             [3, 1; 5, 2],
             [4, 0; 4, 3],
             [4, 1; 4, 2],
             [4, 1; 3, 3]]
            sage: SuperPartition([[2,1],[3]]).add_horizontal_border_strip_star(2)
            [[2, 1; 5], [2, 0; 4, 2], [2, 1; 4, 1], [2, 0; 3, 3], [2, 1; 3, 2]]
        """
        sp1, circ_list = self.to_circled_diagram()
        nsp = [list(la) + [0] for la in sp1.add_horizontal_border_strip(h)]
        sp1 = sp1 + [0]
        out = []
        for elt in nsp:
            row_changed = [row1 - row2 for row1, row2 in zip(elt, sp1)]
            new_sp = [elt, [(i[0] + 1, elt[i[0] + 1]) for i in circ_list
                            if row_changed[i[0]] != 0]
                      # TODO: Check that this is not supposed to be
                      #   a tuple of size 1
                      + [(i) for i in circ_list if row_changed[i[0]] == 0]]
            if len({k for j, k in new_sp[1]}) == len(new_sp[1]):
                out += [SuperPartition.from_circled_diagram(*new_sp)]
        return out

    def add_horizontal_border_strip_star_bar(self, h) -> list:
        r"""
        List super partitions that differ from ``self`` by a horizontal strip.

        The notion of horizontal strip comes from the Pieri rule for the
        Schur-star-bar basis of symmetric functions in super space (see
        Theorem 10 from [JL2016]_).

        INPUT:

        - ``h`` -- number of cells in the horizontal strip

        OUTPUT: list of super partitions

        EXAMPLES::

            sage: SuperPartition([[4,1],[5,4]]).add_horizontal_border_strip_star_bar(3)
            [[4, 1; 8, 4],
             [4, 1; 7, 5],
             [4, 2; 7, 4],
             [4, 1; 7, 4, 1],
             [4, 2; 6, 5],
             [4, 1; 6, 5, 1],
             [4, 3; 6, 4],
             [4, 2; 6, 4, 1],
             [4, 1; 6, 4, 2],
             [4, 3; 5, 5],
             [4, 2; 5, 5, 1],
             [4, 1; 5, 5, 2],
             [4, 3; 5, 4, 1],
             [4, 1; 5, 4, 3]]
            sage: SuperPartition([[3,1],[5]]).add_horizontal_border_strip_star_bar(2)
            [[3, 1; 7],
             [4, 1; 6],
             [3, 2; 6],
             [3, 1; 6, 1],
             [4, 2; 5],
             [4, 1; 5, 1],
             [3, 2; 5, 1],
             [3, 1; 5, 2]]
        """
        sp1, circ_list = self.to_circled_diagram()
        nsp = [list(la) + [0] for la in sp1.add_horizontal_border_strip(h)]
        sp1 = sp1 + [0]
        out = []
        for asp in nsp:
            asp = asp + [0]
            change_in_rows = [asp[i] - sp1[i] for i in range(len(sp1))]
            moved_circ_list: list[list[tuple]] = [[] for _ in range(len(circ_list))]
            for i, pos in enumerate(circ_list):
                if change_in_rows[pos[0]] == 0:
                    moved_circ_list[i].append(pos)
                else:
                    if pos[0] == 0:
                        moved_circ_list[i].append((0, pos[1] + change_in_rows[0]))
                        if pos[1] == asp[1]:
                            moved_circ_list[i].append((1, asp[1]))
                    else:
                        if pos[1] + change_in_rows[pos[0]] < sp1[pos[0] - 1]:
                            moved_circ_list[i].append((pos[0], pos[1] + change_in_rows[pos[0]]))
                        if asp[pos[0] + 1] == sp1[pos[0]]:
                            moved_circ_list[i].append((pos[0] + 1, pos[1]))
            out += [[moved_circ_list, asp]]
        result = []
        for ti in out:
            if not ti[0]:
                result += [[ti[1], ti[0]]]
            else:
                x = reduce(lambda a, b: [item_a + item_b for item_a in a for item_b in b], ti[0])
                for j in x:
                    result += [[ti[1], list(zip(j, j[1:]))[::2]]]
        return [SuperPartition.from_circled_diagram(*ti)
                for ti in result if len(ti[1]) == len(self[0])]


class SuperPartitions(UniqueRepresentation, Parent):
    r"""
    Super partitions.

    A super partition of size `n` and fermionic sector `m` is a
    pair consisting of a strict partition of some integer `r` of
    length `m` (that may end in a `0`) and an integer partition of
    `n - r`.

    INPUT:

    - ``n`` -- integer (default: ``None``)
    - ``m`` -- if ``n`` is specified, an integer (optional: default ``None``)

    Super partitions are the indexing set for symmetric functions
    in super space.

    EXAMPLES::

        sage: SuperPartitions()
        Super Partitions
        sage: SuperPartitions(2)
        Super Partitions of 2
        sage: SuperPartitions(2).cardinality()
        8
        sage: SuperPartitions(4,2)
        Super Partitions of 4 and of fermionic sector 2
        sage: [[2,0],[1,1]] in SuperPartitions(4,2)
        True
        sage: [[1,0],[1,1]] in SuperPartitions(4,2)
        False
        sage: [[1,0],[2,1]] in SuperPartitions(4)
        True
        sage: [[1,0],[2,2,1]] in SuperPartitions(4)
        False
        sage: [[1,0],[2,1]] in SuperPartitions()
        True
        sage: [[1,1],[2,1]] in SuperPartitions()
        False
    """
    @staticmethod
    def __classcall_private__(self, n=None, m=None, **kwargs):
        r"""
        Return the corresponding parent based upon input.

        TESTS::

            sage: from sage.combinat.superpartition import *
            sage: isinstance(SuperPartitions(), SuperPartitions_all)
            True
            sage: isinstance(SuperPartitions(3), SuperPartitions_n)
            True
            sage: isinstance(SuperPartitions(3,2), SuperPartitions_n_m)
            True

        ::

            sage: SP = SuperPartitions(5,2)
            sage: SP2 = SuperPartitions(int(5),int(2))
            sage: SP3 = SuperPartitions(ZZ(5),int(2))
            sage: SP is SP2
            True
            sage: SP is SP3
            True

        ::

            sage: SP = SuperPartitions(5)
            sage: SP2 = SuperPartitions(int(5))
            sage: SP3 = SuperPartitions(ZZ(5))
            sage: SP is SP2
            True
            sage: SP is SP3
            True
        """
        if n is None:
            return SuperPartitions_all()
        elif n in ZZ:
            if m is None:
                return SuperPartitions_n(n)
            elif m in ZZ:
                return SuperPartitions_n_m(n, m)
            raise ValueError("m must be an integer")
        raise ValueError("n must be an integer")

    def __init__(self, is_infinite=False):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: SP = SuperPartitions()
            sage: TestSuite(SP).run()
        """
        cat = EnumeratedSets()
        if is_infinite:
            cat = cat.Infinite()
        else:
            cat = cat.Finite()
        Parent.__init__(self, category=cat)

    Element = SuperPartition

    class options(GlobalOptions):
        """
        Set the global options for elements of the SuperPartition class.

        The defaults are for Super Partitions to be displayed in a list
        notation with the fermionic part and the bosonic part separated
        by a semicolon. There is a slight disadvantage to this notation
        because a list containing a semicolon can not be used as input
        for a super partition.

        @OPTIONS@

        EXAMPLES::

            sage: sp = SuperPartition([[1, 0], [2, 2, 1]])
            sage: SuperPartitions.options.display
            default
            sage: sp
            [1, 0; 2, 2, 1]
            sage: SuperPartitions.options.display = 'list'
            sage: sp
            [-1, 0, 2, 2, 1]
            sage: SuperPartitions.options._reset()
        """,
        NAME = 'SuperPartition'
        module = 'sage.combinat.superpartition'
        display = dict(default='default',
                       description="Specifies how the super partitions should "
                                   "be printed",
                       values=dict(list="the super partitions are displayed in "
                                        "a list of two lists",
                                   pair="the super partition is displayed as a "
                                        "list of integers",
                                   default="the super partition is displayed in "
                                           "a form [fermionic part; bosonic part]"),
                       case_sensitive=False)

    def _element_constructor_(self, lst, check=True):
        """
        Construct an element with ``self`` as parent.

        EXAMPLES::

            sage: SP = SuperPartitions()
            sage: SP([[],[3,3,1]])
            [; 3, 3, 1]
            sage: SP([[],[3,3,1]]) in SP
            True
            sage: SP([[],[3,3,1]]).parent()
            Super Partitions
            sage: SuperPartitions(7)([[],[3,3,1]])
            [; 3, 3, 1]
            sage: SuperPartitions(7,0)([[],[3,3,1]])
            [; 3, 3, 1]
            sage: SuperPartitions(7,1)([[],[3,3,1]])
            Traceback (most recent call last):
            ...
            ValueError: [[], [3, 3, 1]] not in Super Partitions of 7 and of fermionic sector 1
        """
        if not lst:
            return self.element_class(self, [[], []], check=check)
        if isinstance(lst, SuperPartition):
            lst = list(lst)
        if isinstance(lst[0], (list, tuple)):
            return self.element_class(self, [lst[0], [a for a in lst[1] if a > 0]],
                                      check=check)
        else:
            return self.element_class(self, [[-a for a in lst if a <= 0],
                                             [a for a in lst if a > 0]],
                                      check=check)

    def __contains__(self, x) -> bool:
        """
        TESTS::

            sage: [[1],[2,1]] in SuperPartitions()
            True
            sage: [[],[]] in SuperPartitions()
            True
            sage: [[0],[]] in SuperPartitions()
            True
            sage: [[],[0]] in SuperPartitions()
            True
            sage: [-1, 2, 1] in SuperPartitions()
            True
            sage: [2, -1, 1, 0] in SuperPartitions()
            True
            sage: [2, 0, 1, -1] in SuperPartitions()
            False
            sage: [] in SuperPartitions()
            True
            sage: [0] in SuperPartitions()
            True
        """
        if isinstance(x, SuperPartition):
            return True
        if not isinstance(x, (list, tuple)):
            return False
        if all(isinstance(i, (int, Integer)) or i in ZZ for i in x):
            sp = [a for a in x if a <= 0]
            return (all(sp[i] > sp[i - 1] for i in range(1, len(sp)))
                    and [a for a in x if a > 0] in _Partitions)
        elif (len(x) == 2 and
              isinstance(x[0], (list, tuple)) and
              isinstance(x[1], (list, tuple))):
            for i in chain(x[0], x[1]):
                if i not in ZZ:
                    return False
                if i < 0:
                    return False
            return (all(x[0][i] > x[0][i + 1] for i in range(len(x[0]) - 1))
                    and all(x[1][i] >= x[1][i + 1] for i in range(len(x[1]) - 1))
                    and ((not x[0]) or x[0][-1] >= 0) and ((not x[1]) or x[1][-1] >= 0))
        else:
            return False


class SuperPartitions_n_m(SuperPartitions):
    def __init__(self, n, m):
        """
        Initialize ``self``.

        TESTS::

            sage: SP = SuperPartitions(3,2)
            sage: TestSuite(SP).run()
        """
        self.n = n
        self.m = m
        SuperPartitions.__init__(self, False)

    def _repr_(self) -> str:
        """
        Return a string representation of ``self``.

        TESTS::

            sage: repr(SuperPartitions(3,2))
            'Super Partitions of 3 and of fermionic sector 2'
        """
        return "Super Partitions of %s and of fermionic sector %s" % (self.n, self.m)

    def __contains__(self, x) -> bool:
        """
        TESTS::

            sage: [[3,2,1,0],[2]] in SuperPartitions(8,4)
            True
            sage: [[3,2,1,0],[]] in SuperPartitions(6,3)
            False
            sage: [[],[]] in SuperPartitions(0,0)
            True
            sage: [[0],[]] in SuperPartitions(0,1)
            True
            sage: [[],[]] in SuperPartitions(0,1)
            False
            sage: [-3,-2,-1,0,2] in SuperPartitions(8,4)
            True
            sage: [0] in SuperPartitions(0,0)
            False
            sage: [] in SuperPartitions(0,0)
            True
            sage: [0] in SuperPartitions(0,1)
            True
        """
        if x in SuperPartitions():
            if not x:
                return self.n == 0 and self.m == 0
            if isinstance(x[0], (list, tuple)):
                n = sum(x[0] + x[1])
                m = len(x[0])
            else:
                n = sum(abs(a) for a in x)
                m = len([a for a in x if a <= 0])
            return n == self.n and m == self.m
        else:
            return False

    def __iter__(self):
        r"""
        An iterator for super partitions of degree ``n`` and sector ``m``.

        EXAMPLES::

            sage: SuperPartitions(6,2).cardinality()
            28
            sage: SuperPartitions(6,4).first()
            [3, 2, 1, 0; ]
        """
        for r in range(self.n + 1):
            for p1 in Partitions(r):
                for p0 in Partitions(self.n - r, max_slope=-1, length=self.m):
                    yield self.element_class(self, [list(p0), list(p1)])
                for p0 in Partitions(self.n - r, max_slope=-1, length=self.m - 1):
                    yield self.element_class(self, [list(p0) + [0], list(p1)])


class SuperPartitions_n(SuperPartitions):
    def __init__(self, n):
        """
        Initialize ``self``.

        TESTS::

            sage: SP = SuperPartitions(3)
            sage: TestSuite(SP).run()
        """
        self.n = n
        SuperPartitions.__init__(self, False)

    def _repr_(self) -> str:
        """
        Return a string representation of ``self``.

        TESTS::

            sage: repr(SuperPartitions(3))
            'Super Partitions of 3'
        """
        return "Super Partitions of %s" % self.n

    def __contains__(self, x) -> bool:
        """
        EXAMPLES::

            sage: SuperPartitions(7)([[],[3,3,1]]) in SuperPartitions()
            True
            sage: SuperPartitions()([[],[3,3,1]]) in SuperPartitions(7)
            True
            sage: [[],[]] in SuperPartitions(0)
            True
            sage: [[0],[]] in SuperPartitions(0)
            True
            sage: [0] in SuperPartitions(0)
            True
            sage: [] in SuperPartitions(0)
            True
            sage: [1] in SuperPartitions(0)
            False
        """
        if x in SuperPartitions():
            if not x:
                return self.n == 0
            if isinstance(x[0], (list, tuple)):
                n = sum(x[0] + x[1])
            else:
                n = sum(abs(a) for a in x)
            return n == self.n
        else:
            return False

    def __iter__(self):
        r"""
        An iterator for super partitions of degree ``n``.

        EXAMPLES::

            sage: SuperPartitions(1).list()
            [[; 1], [1; ], [0; 1], [1, 0; ]]
            sage: SuperPartitions(6).cardinality()
            80
        """
        m = 0
        while self.n >= m * (m - 1) // 2:
            for LA in SuperPartitions(self.n, m):
                yield self.element_class(self, LA)
            m += 1


class SuperPartitions_all(SuperPartitions):
    def __init__(self):
        """
        Initialize ``self``.

        TESTS::

            sage: SP = SuperPartitions()
            sage: TestSuite(SP).run()
        """
        SuperPartitions.__init__(self, True)

    def _repr_(self) -> str:
        """
        Return a string representation of ``self``.

        TESTS::

            sage: repr(SuperPartitions())
            'Super Partitions'
        """
        return "Super Partitions"

    def __iter__(self):
        """
        Iterate over all super partitions.

        EXAMPLES::

            sage: SP = SuperPartitions()
            sage: it = SP.__iter__()
            sage: [next(it) for i in range(6)]
            [[; ], [0; ], [; 1], [1; ], [0; 1], [1, 0; ]]
        """
        n = 0
        while True:
            for sp in SuperPartitions(n):
                yield self.element_class(self, list(sp))
            n += 1
